// DigitalInputs.js
//  This module handles digital input processing and eventing

//#region Dependencies, Constants and Globals
//#region Dependencies
var fs = require('fs');
var moment = require('moment');
var exec = require('child_process').exec;
//#endregion Dependencies

//#region Constants
// Define the configuraton for the I/O
var DIGITAL_INPUTS_LOCATION = '/sys/class/gpio/gpio';
var DIGITAL_INPUTS_BASEGPIO = 448;// Default=448
var DIGITAL_INPUTS_POINTS = 48;// Default=48

// SNMP - this is here for now, could move
var AuthProtocol = ["None", "MD5", "SHA"];
var PrivProtocol = ["None", "DES", "AES"];
//#endregion Constants

//#region Globals
// The following array holds, in memory, the current value (as 0 or 1 values)
var digitalInputsArray = [
    { "Address": "0", "Name": "Name", "Value": 0, "TempValue": 0, "ValueChanged": false, "HoldTimer": 0, "Enabled": "FALSE", "SendEvents": "FALSE", "Polarity": "Normally Open", "HoldTime": 0, "TimeChanged": 0 }, //376
    { "Address": "1", "Name": "Name", "Value": 0, "TempValue": 0, "ValueChanged": false, "HoldTimer": 0, "Enabled": "FALSE", "SendEvents": "FALSE", "Polarity": "Normally Open", "HoldTime": 0, "TimeChanged": 0 },
    { "Address": "2", "Name": "Name", "Value": 0, "TempValue": 0, "ValueChanged": false, "HoldTimer": 0, "Enabled": "FALSE", "SendEvents": "FALSE", "Polarity": "Normally Open", "HoldTime": 0, "TimeChanged": 0 },
    { "Address": "3", "Name": "Name", "Value": 0, "TempValue": 0, "ValueChanged": false, "HoldTimer": 0, "Enabled": "FALSE", "SendEvents": "FALSE", "Polarity": "Normally Open", "HoldTime": 0, "TimeChanged": 0 },
    { "Address": "4", "Name": "Name", "Value": 0, "TempValue": 0, "ValueChanged": false, "HoldTimer": 0, "Enabled": "FALSE", "SendEvents": "FALSE", "Polarity": "Normally Open", "HoldTime": 0, "TimeChanged": 0 },
    { "Address": "5", "Name": "Name", "Value": 0, "TempValue": 0, "ValueChanged": false, "HoldTimer": 0, "Enabled": "FALSE", "SendEvents": "FALSE", "Polarity": "Normally Open", "HoldTime": 0, "TimeChanged": 0 },
    { "Address": "6", "Name": "Name", "Value": 0, "TempValue": 0, "ValueChanged": false, "HoldTimer": 0, "Enabled": "FALSE", "SendEvents": "FALSE", "Polarity": "Normally Open", "HoldTime": 0, "TimeChanged": 0 },
    { "Address": "7", "Name": "Name", "Value": 0, "TempValue": 0, "ValueChanged": false, "HoldTimer": 0, "Enabled": "FALSE", "SendEvents": "FALSE", "Polarity": "Normally Open", "HoldTime": 0, "TimeChanged": 0 },

    { "Address": "8", "Name": "Name", "Value": 0, "TempValue": 0, "ValueChanged": false, "HoldTimer": 0, "Enabled": "FALSE", "SendEvents": "FALSE", "Polarity": "Normally Open", "HoldTime": 0, "TimeChanged": 0 },
    { "Address": "9", "Name": "Name", "Value": 0, "TempValue": 0, "ValueChanged": false, "HoldTimer": 0, "Enabled": "FALSE", "SendEvents": "FALSE", "Polarity": "Normally Open", "HoldTime": 0, "TimeChanged": 0 },
    { "Address": "10", "Name": "Name", "Value": 0, "TempValue": 0, "ValueChanged": false, "HoldTimer": 0, "Enabled": "FALSE", "SendEvents": "FALSE", "Polarity": "Normally Open", "HoldTime": 0, "TimeChanged": 0 },
    { "Address": "11", "Name": "Name", "Value": 0, "TempValue": 0, "ValueChanged": false, "HoldTimer": 0, "Enabled": "FALSE", "SendEvents": "FALSE", "Polarity": "Normally Open", "HoldTime": 0, "TimeChanged": 0 },
    { "Address": "12", "Name": "Name", "Value": 0, "TempValue": 0, "ValueChanged": false, "HoldTimer": 0, "Enabled": "FALSE", "SendEvents": "FALSE", "Polarity": "Normally Open", "HoldTime": 0, "TimeChanged": 0 },
    { "Address": "13", "Name": "Name", "Value": 0, "TempValue": 0, "ValueChanged": false, "HoldTimer": 0, "Enabled": "FALSE", "SendEvents": "FALSE", "Polarity": "Normally Open", "HoldTime": 0, "TimeChanged": 0 },
    { "Address": "14", "Name": "Name", "Value": 0, "TempValue": 0, "ValueChanged": false, "HoldTimer": 0, "Enabled": "FALSE", "SendEvents": "FALSE", "Polarity": "Normally Open", "HoldTime": 0, "TimeChanged": 0 },
    { "Address": "15", "Name": "Name", "Value": 0, "TempValue": 0, "ValueChanged": false, "HoldTimer": 0, "Enabled": "FALSE", "SendEvents": "FALSE", "Polarity": "Normally Open", "HoldTime": 0, "TimeChanged": 0 },

    { "Address": "16", "Name": "Name", "Value": 0, "TempValue": 0, "ValueChanged": false, "HoldTimer": 0, "Enabled": "FALSE", "SendEvents": "FALSE", "Polarity": "Normally Open", "HoldTime": 0, "TimeChanged": 0 },
    { "Address": "17", "Name": "Name", "Value": 0, "TempValue": 0, "ValueChanged": false, "HoldTimer": 0, "Enabled": "FALSE", "SendEvents": "FALSE", "Polarity": "Normally Open", "HoldTime": 0, "TimeChanged": 0 },
    { "Address": "18", "Name": "Name", "Value": 0, "TempValue": 0, "ValueChanged": false, "HoldTimer": 0, "Enabled": "FALSE", "SendEvents": "FALSE", "Polarity": "Normally Open", "HoldTime": 0, "TimeChanged": 0 },
    { "Address": "19", "Name": "Name", "Value": 0, "TempValue": 0, "ValueChanged": false, "HoldTimer": 0, "Enabled": "FALSE", "SendEvents": "FALSE", "Polarity": "Normally Open", "HoldTime": 0, "TimeChanged": 0 },
    { "Address": "20", "Name": "Name", "Value": 0, "TempValue": 0, "ValueChanged": false, "HoldTimer": 0, "Enabled": "FALSE", "SendEvents": "FALSE", "Polarity": "Normally Open", "HoldTime": 0, "TimeChanged": 0 },
    { "Address": "21", "Name": "Name", "Value": 0, "TempValue": 0, "ValueChanged": false, "HoldTimer": 0, "Enabled": "FALSE", "SendEvents": "FALSE", "Polarity": "Normally Open", "HoldTime": 0, "TimeChanged": 0 },
    { "Address": "22", "Name": "Name", "Value": 0, "TempValue": 0, "ValueChanged": false, "HoldTimer": 0, "Enabled": "FALSE", "SendEvents": "FALSE", "Polarity": "Normally Open", "HoldTime": 0, "TimeChanged": 0 },
    { "Address": "23", "Name": "Name", "Value": 0, "TempValue": 0, "ValueChanged": false, "HoldTimer": 0, "Enabled": "FALSE", "SendEvents": "FALSE", "Polarity": "Normally Open", "HoldTime": 0, "TimeChanged": 0 },

    { "Address": "24", "Name": "Name", "Value": 0, "TempValue": 0, "ValueChanged": false, "HoldTimer": 0, "Enabled": "FALSE", "SendEvents": "FALSE", "Polarity": "Normally Open", "HoldTime": 0, "TimeChanged": 0 }, //352
    { "Address": "25", "Name": "Name", "Value": 0, "TempValue": 0, "ValueChanged": false, "HoldTimer": 0, "Enabled": "FALSE", "SendEvents": "FALSE", "Polarity": "Normally Open", "HoldTime": 0, "TimeChanged": 0 },
    { "Address": "26", "Name": "Name", "Value": 0, "TempValue": 0, "ValueChanged": false, "HoldTimer": 0, "Enabled": "FALSE", "SendEvents": "FALSE", "Polarity": "Normally Open", "HoldTime": 0, "TimeChanged": 0 },
    { "Address": "27", "Name": "Name", "Value": 0, "TempValue": 0, "ValueChanged": false, "HoldTimer": 0, "Enabled": "FALSE", "SendEvents": "FALSE", "Polarity": "Normally Open", "HoldTime": 0, "TimeChanged": 0 },
    { "Address": "28", "Name": "Name", "Value": 0, "TempValue": 0, "ValueChanged": false, "HoldTimer": 0, "Enabled": "FALSE", "SendEvents": "FALSE", "Polarity": "Normally Open", "HoldTime": 0, "TimeChanged": 0 },
    { "Address": "29", "Name": "Name", "Value": 0, "TempValue": 0, "ValueChanged": false, "HoldTimer": 0, "Enabled": "FALSE", "SendEvents": "FALSE", "Polarity": "Normally Open", "HoldTime": 0, "TimeChanged": 0 },
    { "Address": "30", "Name": "Name", "Value": 0, "TempValue": 0, "ValueChanged": false, "HoldTimer": 0, "Enabled": "FALSE", "SendEvents": "FALSE", "Polarity": "Normally Open", "HoldTime": 0, "TimeChanged": 0 },
    { "Address": "31", "Name": "Name", "Value": 0, "TempValue": 0, "ValueChanged": false, "HoldTimer": 0, "Enabled": "FALSE", "SendEvents": "FALSE", "Polarity": "Normally Open", "HoldTime": 0, "TimeChanged": 0 },

    { "Address": "32", "Name": "Name", "Value": 0, "TempValue": 0, "ValueChanged": false, "HoldTimer": 0, "Enabled": "FALSE", "SendEvents": "FALSE", "Polarity": "Normally Open", "HoldTime": 0, "TimeChanged": 0 },
    { "Address": "33", "Name": "Name", "Value": 0, "TempValue": 0, "ValueChanged": false, "HoldTimer": 0, "Enabled": "FALSE", "SendEvents": "FALSE", "Polarity": "Normally Open", "HoldTime": 0, "TimeChanged": 0 },
    { "Address": "34", "Name": "Name", "Value": 0, "TempValue": 0, "ValueChanged": false, "HoldTimer": 0, "Enabled": "FALSE", "SendEvents": "FALSE", "Polarity": "Normally Open", "HoldTime": 0, "TimeChanged": 0 },
    { "Address": "35", "Name": "Name", "Value": 0, "TempValue": 0, "ValueChanged": false, "HoldTimer": 0, "Enabled": "FALSE", "SendEvents": "FALSE", "Polarity": "Normally Open", "HoldTime": 0, "TimeChanged": 0 },
    { "Address": "36", "Name": "Name", "Value": 0, "TempValue": 0, "ValueChanged": false, "HoldTimer": 0, "Enabled": "FALSE", "SendEvents": "FALSE", "Polarity": "Normally Open", "HoldTime": 0, "TimeChanged": 0 },
    { "Address": "37", "Name": "Name", "Value": 0, "TempValue": 0, "ValueChanged": false, "HoldTimer": 0, "Enabled": "FALSE", "SendEvents": "FALSE", "Polarity": "Normally Open", "HoldTime": 0, "TimeChanged": 0 },
    { "Address": "38", "Name": "Name", "Value": 0, "TempValue": 0, "ValueChanged": false, "HoldTimer": 0, "Enabled": "FALSE", "SendEvents": "FALSE", "Polarity": "Normally Open", "HoldTime": 0, "TimeChanged": 0 },
    { "Address": "39", "Name": "Name", "Value": 0, "TempValue": 0, "ValueChanged": false, "HoldTimer": 0, "Enabled": "FALSE", "SendEvents": "FALSE", "Polarity": "Normally Open", "HoldTime": 0, "TimeChanged": 0 },

    { "Address": "40", "Name": "Name", "Value": 0, "TempValue": 0, "ValueChanged": false, "HoldTimer": 0, "Enabled": "FALSE", "SendEvents": "FALSE", "Polarity": "Normally Open", "HoldTime": 0, "TimeChanged": 0 },
    { "Address": "41", "Name": "Name", "Value": 0, "TempValue": 0, "ValueChanged": false, "HoldTimer": 0, "Enabled": "FALSE", "SendEvents": "FALSE", "Polarity": "Normally Open", "HoldTime": 0, "TimeChanged": 0 },
    { "Address": "42", "Name": "Name", "Value": 0, "TempValue": 0, "ValueChanged": false, "HoldTimer": 0, "Enabled": "FALSE", "SendEvents": "FALSE", "Polarity": "Normally Open", "HoldTime": 0, "TimeChanged": 0 },
    { "Address": "43", "Name": "Name", "Value": 0, "TempValue": 0, "ValueChanged": false, "HoldTimer": 0, "Enabled": "FALSE", "SendEvents": "FALSE", "Polarity": "Normally Open", "HoldTime": 0, "TimeChanged": 0 },
    { "Address": "44", "Name": "Name", "Value": 0, "TempValue": 0, "ValueChanged": false, "HoldTimer": 0, "Enabled": "FALSE", "SendEvents": "FALSE", "Polarity": "Normally Open", "HoldTime": 0, "TimeChanged": 0 },
    { "Address": "45", "Name": "Name", "Value": 0, "TempValue": 0, "ValueChanged": false, "HoldTimer": 0, "Enabled": "FALSE", "SendEvents": "FALSE", "Polarity": "Normally Open", "HoldTime": 0, "TimeChanged": 0 },
    { "Address": "46", "Name": "Name", "Value": 0, "TempValue": 0, "ValueChanged": false, "HoldTimer": 0, "Enabled": "FALSE", "SendEvents": "FALSE", "Polarity": "Normally Open", "HoldTime": 0, "TimeChanged": 0 },
    { "Address": "47", "Name": "Name", "Value": 0, "TempValue": 0, "ValueChanged": false, "HoldTimer": 0, "Enabled": "FALSE", "SendEvents": "FALSE", "Polarity": "Normally Open", "HoldTime": 0, "TimeChanged": 0 }
];

var SummitInfo = {};
var SmtpManager = {};
var SnmpManager = {};
var SqliteManager = {};
var LogManager = {};
var SocketIoManager = {};
//#endregion Globals
//#endregion Dependencies, Constants and Globals

//#region Digital Inputs Initialization
//      - Checks if database table exists, if not creates the table and fills it with default data for each point
//      - Else reads from the database and updates digitalInputsArray
exports.InitializeDigitalInputs = function (info, smtp, snmp, db, io, logger, callback) {
    // Initialize globals
    SummitInfo = info;
    SmtpManager = smtp;
    SnmpManager = snmp;
    SqliteManager = db;
    SocketIoManager = io;
    LogManager = logger;

    // Try getting the DigitalInputs table
    SqliteManager.get("SELECT name FROM sqlite_master WHERE type='table' AND name='DigitalInputs'", function (err, rows) {
        if (err !== null) {
            LogManager.error('Digital Inputs - Initialize Error (Check)=' + JSON.stringify(err));
        }
        else if (typeof rows === "undefined") {
            // We read nothing, so let's create the table ...
            LogManager.info("Digital Inputs - Initialize by creating database table ...");
            SqliteManager.exec("CREATE TABLE if not exists DigitalInputs (Address INTEGER, Name TEXT, Value TEXT, Enabled TEXT, SendEvents TEXT, Polarity TEXT, HoldTime INTEGER, TimeChanged TEXT )", function (err) {
                if (err !== null) {
                    LogManager.error('Digital Inputs - Initialize Error (Create)=' + JSON.stringify(err));
                }
                else {
                    var timeChanged = moment();// Record the create time as the last time the I/O point value was changed

                    // Created the table, now add data ...
                    SqliteManager.run("BEGIN")
                    for (var i = 0; i < DIGITAL_INPUTS_POINTS; i++) {
                        // Initialize the array
                        UpdateDigitalInputConfiguration(i, "Digital Input " + (i + 1), 0, "FALSE", "FALSE", "Normally Open", 0, timeChanged);
                        digitalInputsArray[i].TempValue = -1;
                        digitalInputsArray[i].ValueChanged = false;
                        digitalInputsArray[i].HoldTimer = timeChanged;

                        // Initialize the database
                        insertStatement = i + ", 'Digital Input " + (i + 1) + "', '0', 'FALSE', 'FALSE', 'Normally Open', 0, '" + timeChanged + "'";
                        SqliteManager.run("INSERT INTO DigitalInputs (Address, Name, Value, Enabled, SendEvents, Polarity, HoldTime, TimeChanged) VALUES (" + insertStatement + ")", function (err) {
                            if (err !== null) {
                                LogManager.error('Digital Inputs - Initialize Error (Insert)=' + JSON.stringify(err));
                            }
                        });
                    }
                    SqliteManager.run("COMMIT");

                    // Digital Inputs table created
                    if (typeof callback == "function") {
                        LogManager.info('Digital Inputs - Initialized new database table');
                        callback();
                    }
                }
            });
        }
        else {
            // Read from the database and populate the digitalInputsArray
            LogManager.info("Digital Inputs - Initialize by reading database ...");
            SqliteManager.each("SELECT rowid as id, Address, Name, Value, Enabled, SendEvents, Polarity, HoldTime, TimeChanged FROM DigitalInputs ORDER BY Address", function (err, row) {
                if (err !== null) {
                    LogManager.error('Digital Inputs - Initialize Error (Select)=' + JSON.stringify(err));
                }
                else {
                    UpdateDigitalInputConfiguration(row.Address, row.Name, row.Value, row.Enabled, row.SendEvents, row.Polarity, row.HoldTime, row.TimeChanged);
                    LogManager.debug("Digital Inputs - [Value INITIAL from dB] Digital Input[" + (parseInt(row.Address) + 1) + "] '" + row.Name + "' Value=" + row.Value);
                    digitalInputsArray[row.Address].TempValue = row.Value;
                    digitalInputsArray[row.Address].HoldTimer = 0;// This should be 0 to start
                }
            }, function () {
                // Digital Inputs read from table
                if (typeof callback == "function") {
                    LogManager.info('Digital Inputs - Initialized from existing database table');
                    callback();
                }
            });
        }
    });
}
//#endregion Digital Inputs Initialization

//#region Digital Inputs API
module.exports.UpdateDigitalInputConfiguration = UpdateDigitalInputConfiguration;
function UpdateDigitalInputConfiguration(Address, Name, Value, Enabled, SendEvents, Polarity, HoldTime, TimeChanged) {
    digitalInputsArray[Address].Name = Name;
    digitalInputsArray[Address].Value = Value;
    digitalInputsArray[Address].Enabled = Enabled;
    digitalInputsArray[Address].SendEvents = SendEvents;
    digitalInputsArray[Address].Polarity = Polarity;
    digitalInputsArray[Address].HoldTime = HoldTime;
    digitalInputsArray[Address].TimeChanged = TimeChanged;
    sendValueDigitalInputs();// Update the Digital Inputs configuration
}
//#endregion Digital Inputs API

//#region Digital Inputs Processing
var processCounter = 0;
var processTotalTime = 0;
exports.ProcessDigitalInputs = function (ledManager, callback) {
    timeStart = moment();

    var index;// This is an index to digital I/O point - accounts for first 24 bits being at a higher location that the second 24 bits
    var fileIndex;// This is an index to the GPIO location - it's the index + the GPIO offset
    var offset = 0;
    var timeChanged = moment();
    try {
        for (index = 0; index < DIGITAL_INPUTS_POINTS; index++) {
            // Figure file to read input value
            //#region File name logic
            // This f'ed up logic is that first 24 I/O points are higher than the next 24 I/O points
            if (offset < 24) {
                fileIndex = DIGITAL_INPUTS_BASEGPIO + index + 24;
            }
            else {
                fileIndex = DIGITAL_INPUTS_BASEGPIO + index - 24;
            }
            offset++;
            //#endregion File name logic

            // Only read value if the I/O point is enabled
            if (digitalInputsArray[index].Enabled === "TRUE") {
                // Read raw value from file
                //  If the raw value has changed then we'll ahve to do stuff
                //#region Read raw value
                data = fs.readFileSync(DIGITAL_INPUTS_LOCATION + fileIndex + "/value");
                readValue = String(data).replace(/(\r\n|\n|\r)/gm, "");

                // Normally Closed vs Normally Open
                if (digitalInputsArray[index].Polarity == "Normally Open") {
                    if (readValue === "1") {
                        readValue = "0";
                    } else {
                        readValue = "1";
                    }
                }
                //#endregion Read raw value

                if ((digitalInputsArray[index].Value != readValue) || (digitalInputsArray[index].HoldTime != 0)) {
                    // Filter logic
                    //  - We ignore non changes in digital values
                    //  - But we need to go through the logic when a hold timer is being used
                    //#region Filter logic                  
                    if (digitalInputsArray[index].HoldTime > 0) {
                        LogManager.trace("Digital Inputs - [Filter: HoldTimer Used(" + digitalInputsArray[index].HoldTime + ")] Digital Input[" + (index + 1) + "] old value[array]=" + digitalInputsArray[index].Value + " new value[read]=" + readValue);
                    } else {
                        LogManager.trace("Digital Inputs - [Filter:  Value Changed] Digital Input[" + (index + 1) + "] old value[array]=" + digitalInputsArray[index].Value + " new value[read]=" + readValue);
                    }
                    //#endregion Filter logic

                    // HoldTime logic
                    //  - A non-zero value in the HoldTimer variable means that the HoldTimer is active
                    //  - Start the timer if there was a change and the timer wasn't previously started
                    //  - If the timer was previously started, then check if the 
                    //#region HoldTime logic
                    var HoldTimeValid = false;
                    var HoldTimeChanged = 0;
                    if (digitalInputsArray[index].HoldTime > 0) {
                        // We are using holdtimes ...
                        if (digitalInputsArray[index].Value != readValue) {
                            // Value has changed ...
                            if (digitalInputsArray[index].HoldTimer === 0) {
                                // HoldTimer not started, so start it
                                digitalInputsArray[index].HoldTimer = timeChanged;// Start the timer
                                LogManager.info("Digital Inputs - [HOLD TIMER - HoldTimer=" + Number(digitalInputsArray[index].HoldTime).toFixed(2) + "s, Timer=STARTED, Valid=FALSE] Digital Input[" + (index + 1) + "] '" + digitalInputsArray[index].Name + "'");
                                SocketIoManager.sockets.emit("DigitalInputHoldTime", { "DigitalInput": index, "HoldTime": "HOLDING" });
                            } else {
                                HoldTimeChanged = (timeChanged - digitalInputsArray[index].HoldTimer) / 1000;
                                if (HoldTimeChanged >= digitalInputsArray[index].HoldTime) {
                                    LogManager.info("Digital Inputs - [HOLD TIMER - HoldTimer=" + Number(digitalInputsArray[index].HoldTime).toFixed(2) + "s, Timer=" + Number(HoldTimeChanged).toFixed(2) + ", HoldTimeValid=TRUE] Digital Input[" + (index + 1) + "] '" + digitalInputsArray[index].Name + "'");
                                    SocketIoManager.sockets.emit("DigitalInputHoldTime", { "DigitalInput": index, "HoldTime": "ON" })
                                    HoldTimeValid = true;
                                    (function (index) {
                                        setTimeout(function () {
                                            SocketIoManager.sockets.emit("DigitalInputHoldTime", { "DigitalInput": index, "HoldTime": "OFF" })
                                        }, 1000);
                                    })(index);
                                    digitalInputsArray[index].HoldTimer = 0;// Set hold timer to 0
                                } else {
                                    LogManager.info("Digital Inputs - [HOLD TIMER - HoldTimer=" + Number(digitalInputsArray[index].HoldTime).toFixed(2) + "s, Timer=" + Number(HoldTimeChanged).toFixed(2) + ", HoldTimeValid=FALSE] Digital Input[" + (index + 1) + "] '" + digitalInputsArray[index].Name + "'");
                                    SocketIoManager.sockets.emit("DigitalInputHoldTime", { "DigitalInput": index, "HoldTime": "HOLDING" })
                                }
                            }
                        } else {
                            // Value has either returned to the same value or hasn't changed, so reset the hold timer
                            if (digitalInputsArray[index].HoldTimer !== 0) {
                                SocketIoManager.sockets.emit("DigitalInputHoldTime", { "DigitalInput": index, "HoldTime": "OFF" })
                                LogManager.trace("Digital Inputs - [HOLD TIMER - HoldTimer=" + Number(digitalInputsArray[index].HoldTime).toFixed(2) + "s, Timer=RESET, HoldTimeValid=FALSE] Digital Input[" + (index + 1) + "] '" + digitalInputsArray[index].Name + "'");
                                digitalInputsArray[index].HoldTimer = 0;
                            }
                        }
                    } else {
                        // Not using a hold time ...
                        HoldTimeValid = true;
                        LogManager.trace("Digital Inputs - [HoldTimer=NOT_USED] Digital Input[" + (index + 1) + "] '" + digitalInputsArray[index].Name + "'");
                    }
                    //#endregion HoldTime logic

                    // Publish logic
                    //  - If the value has changed and the hold timer logic has been met, then change the value            
                    //#region Publish logic
                    if ((digitalInputsArray[index].Value != readValue) && HoldTimeValid) {

                        // Change is valid ... 
                        digitalInputsArray[index].Value = readValue;
                        digitalInputsArray[index].TimeChanged = timeChanged;

                        // Update real-time value
                        sendSocketSnmpSmtp(digitalInputsArray[index])

                        // Update stored value
                        SqliteManager.run("UPDATE DigitalInputs SET Value='" + digitalInputsArray[index].Value
                            + "', TimeChanged='" + digitalInputsArray[index].TimeChanged
                            + "' WHERE Address = '" + index + "'",
                            function () {
                            });

                        LogManager.info("Digital Inputs - [THRESHOLD UPDATE] Digital Input[" + (index + 1) + "] '" + digitalInputsArray[index].Name + "' Value=" + digitalInputsArray[index].Value + " HoldTime=" + (digitalInputsArray[index].HoldTime == 0 ? "0" : Number(HoldTimeChanged).toFixed(2)) + "seconds");
                        LogManager.debug("Digital Inputs - [DATABASE UPDATE] Digital Input[" + (index + 1) + "] '" + digitalInputsArray[index].Name + "' New Value=" + digitalInputsArray[index].Value);
                        LogManager.trace("Digital Inputs - [DATABASE UPDATE] Digital Input[" + (index + 1) + "] =" + JSON.stringify(digitalInputsArray[index]));
                    } else {
                        if (!HoldTimeValid) {
                            //var TempDigital = JSON.parse(JSON.stringify(digitalInputsArray[index]));
                            //TempDigital.Value = readValue;
                            //sendValueDigitalInput("DigitalInput", TempDigital, io, logger);

                            LogManager.trace("Digital Inputs - [REAL-TIME UPDATE - Value Changed but HoldTimerValid=FALSE] Digital Input[" + (index + 1) + "] '" + digitalInputsArray[index].Name + "' New value=" + digitalInputsArray[index].Value);
                        }
                    }
                    //#endregion Publish logic
                }// if ((digitalInputsArray[index].Value != readValue) || (digitalInputsArray[index].HoldTime != 0))
            } // if (digitalInputsArray[index].Enabled === "TRUE")
        }//  for (index = 0; index < DIGITAL_INPUTS_POINTS; index++)
    }//  try
    catch (error) {
        LogManager.error('Digital Inputs - Process Error (Read)=' + JSON.stringify(error));
    }
    finally {
        callback(digitalInputsArray);
    }// finally
}
//#endregion Digital Inputs Processing

//#region Digital Inputs Eventing
function sendSocketSnmpSmtp(digitalInput) {
    // SOCKET.IO: Send changed value to all attached clients
    sendValueDigitalInput("DigitalInput", digitalInput);

    // SNMP: Send SNMP Trap/Notification if threshold(s) crossed
    sendSNMP(digitalInput);

    // SMTP: Send SMTP message if threshold(s) crossed
    sendSMTP(digitalInput)
}

function sendValueDigitalInput(dataType, data) {
    SocketIoManager.sockets.emit(dataType, data);
}

function sendValueDigitalInputs() {
    enabled = 0;
    for (i = 0; i < digitalInputsArray.length; i++) {
        if (digitalInputsArray[i].Enabled === "TRUE") enabled++;
    }
    SocketIoManager.sockets.emit("DigitalInputs", { "DigitalInputsTotal": digitalInputsArray.length, "DigitalInputsEnabled": enabled })
}

function sendSNMP(digitalInput) {
    // SNMP definitions - also refer to the TASC-SUMMIT.MIB file for detailed information
    if (digitalInput.Enabled == "FALSE") return;
    if (digitalInput.SendEvents == "FALSE") return;

    address = parseInt(digitalInput.Address) + 1;// To overcome concatenation
    LogManager.trace("Digital Inputs - Sending SNMP Trap for Digital Input[" + address + "] '" + digitalInput.Name + "'");

    // Define the notification
    var snmpCommand;
    var snmpHost;
    var snmpNotification = "";
    snmpNotification += "1.3.6.1.4.1.11902.3.1.1000.1000.0." + (address - 1 + 1000) + " ";
    snmpNotification += "1.3.6.1.4.1.11902.3.1." + (address) + ".0.1 s '" + moment().toString() + "' ";
    snmpNotification += "1.3.6.1.4.1.11902.3.1." + (address) + ".0.2 i " + (parseInt(digitalInput.Address) + 1) + " ";
    snmpNotification += "1.3.6.1.4.1.11902.3.1." + (address) + ".0.3 s '" + digitalInput.Name + "' ";
    snmpNotification += "1.3.6.1.4.1.11902.3.1." + (address) + ".0.4 s '" + ((digitalInput.Value === "1") ? "ON" : "OFF") + "' ";
    snmpNotification += "1.3.6.1.4.1.11902.3.1." + (address) + ".0.5 i " + ((digitalInput.Value === "1") ? 1 : 0) + " ";
    snmpNotification += "1.3.6.1.4.1.11902.3.1." + (address) + ".0.6 i " + ((digitalInput.Polarity === "Normally Closed") ? 1 : 0) + " ";
    snmpNotification += "1.3.6.1.4.1.11902.3.1." + (address) + ".0.7 i " + parseInt(digitalInput.HoldTime) + " ";
    snmpNotification += "1.3.6.1.4.1.11902.3.1." + (address) + ".0.20 s '" + SummitInfo.MACID + "' ";
    snmpNotification += "1.3.6.1.4.1.11902.3.1." + (address) + ".0.21 s '" + SummitInfo.Name + "' ";

    // Send the SNMP notification
    for (var i = 0; i < SnmpManager.length; i++) {
        if (SnmpManager[i].Enabled === "FALSE") continue;

        // Set up the SNMP trap command
        snmpCommand = "snmptrap ";
        if (SnmpManager[i].SnmpVersion == 2) {
            snmpCommand += "-v 2c ";
            snmpCommand += "-c " + SnmpManager[i].Community + " ";
        }
        else if (SnmpManager[i].SnmpVersion == 3) {
            snmpCommand += "-v 3 ";
            snmpCommand += "-e " + SnmpManager[i].Community + " ";
            snmpCommand += "-u " + SnmpManager[i].AuthUser + " ";
            snmpCommand += "-a " + AuthProtocol[SnmpManager[i].AuthProtocol] + " -A " + SnmpManager[i].AuthPassword + " ";
            snmpCommand += "-x " + PrivProtocol[SnmpManager[i].PrivProtocol] + " -X " + SnmpManager[i].PrivPassword + " ";
        } else continue;

        // Send trap to host ...
        snmpHost = SnmpManager[i].IpAddress + " '' ";
        LogManager.info("Digital Inputs - Sending SNMP Event to '" + SnmpManager[i].HostName + "' for Digital Input[" + address + "] '" + digitalInput.Name + "' Value=" + ((digitalInput.Value === "1") ? "ON" : "OFF"));
        LogManager.trace("Digital Inputs - SendSNMP: " + snmpCommand + snmpHost + snmpNotification);
        exec(snmpCommand + snmpHost + snmpNotification, function (error, stdout, stderr) {
            if (error !== null) {
                LogManager.error("Digital Inputs - SendSNMP. " + error);
            } else {
                LogManager.debug("Digital Inputs - SendSNMP. " + stdout);
            }
        });
    }
}

function sendSMTP(digitalInput) {
    if (digitalInput.Enabled == "FALSE") return;
    if (digitalInput.SendEvents == "FALSE") return;
    // NOTE: Decision to send email is decided within SMTP module

    // Set up the SMTP Email message
    var message = [];
    address = parseInt(digitalInput.Address) + 1;// To overcome concatenation
    var event = "'" + digitalInput.Name + "' has changed to status=" + ((digitalInput.Value === "1") ? "ON" : "OFF");
    message += moment().toString() + "\n";
    message += "Summit MACID=" + SummitInfo.MACID + "\n";
    message += "Summit Name=" + SummitInfo.Name + "\n";
    message += "Address=" + (parseInt(digitalInput.Address) + 1) + "\n";
    message += "Name=" + digitalInput.Name + "\n";
    message += "Value=" + ((digitalInput.Value === "1") ? "ON" : "OFF") + "\n";
    message += "Status=" + ((digitalInput.Value === "1") ? 1 : 0) + "\n";
    message += "Polarity=" + ((digitalInput.Polarity === "Normally Closed") ? 1 : 0) + "\n";
    message += "Hold Time=" + (parseInt(digitalInput.HoldTime)) + "\n"

    var ioInfo = "Digital Input[" + address + "] '" + digitalInput.Name + "'";

    LogManager.info("Digital Inputs - Sending SMTP Event for Digital Input[" + address + "] '" + digitalInput.Name + "' Value=" + ((digitalInput.Value === "1") ? "ON" : "OFF"));
    SmtpManager.SendSmtp(message, ioInfo, event, function (result) {
        LogManager.trace("Digital Inputs - SMTP Email Event Response for Digital Input[" + address + "]: " + result.result);
    })
}

exports.UpdateSnmpManager = function (snmp) {
    SnmpManager = snmp;
}

exports.UpdateSummitInfo = function (info) {
    SummitInfo = info;
}
//#endregion Digital Inputs Eventing