﻿// Users.js
//  Initially, this module was used to manage local authentication
//  As of v0.20.x, this module also now also handles: 
//      - RADIUS server authentication
//      - HTTPS web administration and api (primarily handling keys/certs)
//

// Dependencies and Globals
//#region Dependencies and Globals
var rc4a = require('rc4a');
var moment = require('moment');
var fs = require('fs');
var cp = require('child_process');

var SummitPassPhrase = "TotallyAwesomeSuperCool!";// This is our super duper secure phrase
var SummitAuthenticationConfiguration = {};
var SummitLocalAuthenticationConfiguration = {};
var SummitWebServerConfiguration = null;
var SummitLoginAttempts = [];
var SUMMIT_LOGIN_ATTEMPTS = 10;

var Logger;// Logging global

//#endregion Dependencies and Globals

//#region RADIUS Dependencies and Globals
var radius = require('./node-radius/lib/radius');
var dgram = require('dgram');
var util = require('util');
//#endregion RADIUS Dependencies and Globals


// Initialize
//#region Initialize
exports.Initialize = function (logger) {
    Logger = logger;// Globalize

    ReadAuthenticationConfigurationFile();// Read the authentication method/configuration file
    ReadLocalAuthenticationConfigurationFile();// Read the local authentication file, regardless of authentication method
    ReadWebServerConfigurationFile();// Read the web server file (http/https info)
}

function ReadAuthenticationConfigurationFile() {

    try {
        // Try to read the authentication configuration file ...
        SummitAuthenticationConfiguration = require("./authentication.json");
        Logger.info("Users.ReadAuthenticationConfigurationFile: Authentication Config=" + JSON.stringify(SummitAuthenticationConfiguration));
    }
    catch (error) {
        // Something went wrong, so get golden copy from /factory directory
        //  This will also happen on a fresh install of crest because users.json doesn't exist,
        //  and after a factory reset
        Logger.error("Users.ReadAuthenticationConfigurationFile: Unable to find authentication configuration, loading factory default authentication configuration. " + error);
        cp.exec("cp ./factory/authentication.json ./authentication.json", function (error, stdout, stderr) {
            Logger.trace("Users.ReadAuthenticationConfigurationFile: ./factory/authentication.json ./authentication.json - success=" + stdout);
            if (error !== null) {
                Logger.error("Users.ReadAuthenticationConfigurationFile: Error in recovering factory default authentication configuration. Error=" + error);
            } else {
                // Sync the results so the files are made persistent
                cp.exec("sync", function (error, stdout, stderr) {
                    if (error !== null) {
                        Logger.error("Users.ReadAuthenticationConfigurationFile: Error in sync after loading factory default authentication configuration. Error=" + error);
                    } else {
                        SummitAuthenticationConfiguration = require("./authentication.json");
                        Logger.warn("Users.ReadAuthenticationConfigurationFile: Successfully recovered factory default authentication information.");
                    }
                });
            }
        });
    }
}
function ReadLocalAuthenticationConfigurationFile() {

    try {
        // Try to read the users file ...
        SummitLocalAuthenticationConfiguration = require("./users.json");
        Logger.info("Users.ReadLocalAuthenticationConfigurationFile: Local Authentication Config=" + JSON.stringify(SummitLocalAuthenticationConfiguration));
    }
    catch (error) {
        // Something went wrong, so get golden copy from /factory directory
        //  This will also happen on a fresh install of crest because users.json doesn't exist,
        //  and after a factory reset
        Logger.error("Users.GetLocalAuthenticationConfiguration: Unable to find user accounts, loading factory default user accounts. " + error);
        cp.exec("cp ./factory/users.json ./users.json", function (error, stdout, stderr) {
            Logger.trace("Users.LoginUsers: ./factory/users.json ./users.json - success=" + stdout);
            if (error !== null) {
                Logger.error("Users.GetLocalAuthenticationConfiguration: Error in recovering factory default user accounts. Error=" + error);
            } else {
                // Sync the results so the files are made persistent
                cp.exec("sync", function (error, stdout, stderr) {
                    if (error !== null) {
                        Logger.error("Users.GetLocalAuthenticationConfiguration: Error in sync after loading factory default user accounts. Error=" + error);
                    } else {
                        SummitLocalAuthenticationConfiguration = require("./users.json");
                        Logger.warn("Users.GetLocalAuthenticationConfiguration: Successfully recovered factory default local authentication configuration.");
                    }
                });
            }
        });
    }
}
function ReadWebServerConfigurationFile() {

    try {
        // Try to read the web server configuration file ...
        SummitWebServerConfiguration = require("./webserver.json");
        Logger.info("Users.ReadWebServerConfigurationFile: Web Server Config=" + JSON.stringify(SummitAuthenticationConfiguration));
    }
    catch (error) {
        // Something went wrong, so get golden copy from /factory directory
        //  This will also happen on a fresh install of crest because users.json doesn't exist,
        //  and after a factory reset
        Logger.error("Users.ReadWebServerConfigurationFile: Unable to find web server configuration, loading factory default web server configuration. " + error);
        cp.exec("cp ./factory/webserver.json ./webserver.json", function (error, stdout, stderr) {
            Logger.trace("Users.ReadWebServerConfigurationFile: ./factory/webserver.json ./webserver.json - success=" + stdout);
            if (error !== null) {
                Logger.error("Users.ReadWebServerConfigurationFile: Error in recovering factory default web server configuration. Error=" + error);
            } else {
                // Sync the results so the files are made persistent
                cp.exec("sync", function (error, stdout, stderr) {
                    if (error !== null) {
                        Logger.error("Users.ReadWebServerConfigurationFile: Error in sync after loading factory default web server configuration. Error=" + error);
                    } else {
                        SummitWebServerConfiguration = require("./webserver.json");
                        Logger.warn("Users.ReadWebServerConfigurationFile: Successfully recovered factory default web server information.");
                    }
                });
            }
        });
    }
}

//#endregion Initialize

// Authentication: Local and RADIUS
//#region Authentication: Local and RADIUS
//#region Summit Local Authentication Configuration
// Get Summit Local Authentication Configuration
//#region Get Summit Local Authentication Configuration
exports.GetLocalAuthenticationConfiguration = function (callback) {
    // For now, we'll always send decrypted password, so it's easier for client
    //  *** FUTURE: send encrypted password, because client will be able to manage RC4A
    var summitLocalAuthenticationConfiguration = JSON.parse(JSON.stringify(SummitLocalAuthenticationConfiguration));
    //Logger.info("users.GetLocalAuthenticationConfiguration: "+JSON.stringify(SummitLocalAuthenticationConfiguration));
    summitLocalAuthenticationConfiguration.password = rc4a.decrypt(SummitLocalAuthenticationConfiguration.password, SummitPassPhrase);
    callback(summitLocalAuthenticationConfiguration);
}
//#endregion Get Summit Local Authentication Configuration

// Put Summit Local Authentication Configuration
//#region Put Summit Local Authentication Configuration
exports.PutLocalAuthenticationConfiguration = function (summitLocalAuthenticationConfiguration, callback) {
    // For now, we'll always get decrypted password, so it's easier for client
    //  Therefore, we must encrypt before saving
    
    // Update SummitLocalAuthenticationConfiguration
    SummitLocalAuthenticationConfiguration.user = summitLocalAuthenticationConfiguration.user;
    SummitLocalAuthenticationConfiguration.password = rc4a.encrypt(summitLocalAuthenticationConfiguration.password, SummitPassPhrase);
    SummitLocalAuthenticationConfiguration.lastlogin = moment();
    summitLocalAuthenticationConfiguration.lastlogin = SummitLocalAuthenticationConfiguration.lastlogin;
    
    // Update users.json with new SummitLocalAuthenticationConfiguration settings to ensure persistence
    fs.writeFile('users.json', JSON.stringify(SummitLocalAuthenticationConfiguration), function (error) {
        if (error) {
            Logger.error("users.PutSummitUser: User=" + summitLocalAuthenticationConfiguration.user+" account update write error. Error=" + error);
        }
        Logger.info("users.PutSummitUser: User=" + summitLocalAuthenticationConfiguration.user+" account has been udpated");
        callback(summitLocalAuthenticationConfiguration);
    });
}
//#endregion Put Summit Local Authentication Configuration

// GetSummitLoginAttempts
//#region GetSummitLoginAttempts
//  callback    - callback: SummitLoginAttempts
exports.GetSummitLoginAttempts = function (callback) {
    callback(SummitLoginAttempts);
}
//#endregion GetSummitLoginAttempts

// Login User
//  login       - login credentials
//  ip          - IP address of client
//  callback    - callback: success=true, failure=false
//#region Login User
exports.LoginUser = function (login, ip, callback) {
    if (SummitAuthenticationConfiguration.Method == "Local") {
        LocalLogin(login, ip, function (Response) {
            callback(Response);
        });
    }
    else if (SummitAuthenticationConfiguration.Method == "RADIUS Server") {
        RadiusLogin(login, ip, function (Response) {
            callback(Response);
        });
    }
}

function LocalLogin(login, ip, callback) {
    // Response back to client
    var Response = {
        Result: false,
        Message: ""
    };
    var loginTime = moment();

    if ((login.User === SummitLocalAuthenticationConfiguration.user) && (login.Password === rc4a.decrypt(SummitLocalAuthenticationConfiguration.password, SummitPassPhrase))) {

        // Login was successful, save the login time
        SummitLocalAuthenticationConfiguration.lastlogin = loginTime;

        // Save the users file with new login time
        // Update ports.json to ensure persistence
        fs.writeFile('users.json', JSON.stringify(SummitLocalAuthenticationConfiguration), function (error) {
            if (error) {
                Logger.error("Users.LocalLogin: Error writing users account after successful login User='" + login.User + "'. Error=" + error);
                SummitLoginAttempts.unshift({ "time": loginTime, "status": "Internal Error", "user": login.User, "ip": ip });
                if (SummitLoginAttempts.length > SUMMIT_LOGIN_ATTEMPTS) SummitLoginAttempts.splice(-1);

                Response.Message = "Failed authentication configuration update";
            } else {
                Logger.info("Users.LocalLogin: Successful login User='" + login.User + "' (from:" + ip + ") has successfully logged in at " + SummitLocalAuthenticationConfiguration.lastlogin);
                SummitLoginAttempts.unshift({ "time": loginTime, "status": "Successful Login", "user": login.User, "ip": ip });
                if (SummitLoginAttempts.length > SUMMIT_LOGIN_ATTEMPTS) SummitLoginAttempts.splice(-1);

                Response.Result = true;
                Response.Message = "Successful authentication";
            }
            callback(Response);
        });
    }
    else {
        // Credentials do not match ...
        Logger.error("Users.LocalLogin: Failed authentication User='" + login.User + "' (from:" + ip + ") at " + loginTime);
        SummitLoginAttempts.unshift({ "time": loginTime, "status": "Credentials did not match", "user": login.User, "ip": ip });
        if (SummitLoginAttempts.length > SUMMIT_LOGIN_ATTEMPTS) SummitLoginAttempts.splice(-1);

        Response.Message = "Failed authentication";
        callback(Response);
    }
}

function RadiusLogin(login, ip, callback) {
    // Response back to client
    var Response = {
        Result: false,
        Message: ""
    };
    var loginTime = moment();
    var RadiusResponseCount = 0;
    var RadiusPacket = {};

    var LoginAccessRequest = {
        code: "Access-Request",
        secret: SummitAuthenticationConfiguration.RADIUS.Secret,
        identifier: 0,
        attributes: [
            ['NAS-IP-Address', SummitAuthenticationConfiguration.RADIUS.Address],
            ['User-Name', login.User],
            ['User-Password', login.Password]
        ]
    };

    var encoded = radius.encode(LoginAccessRequest);
    RadiusPacket[LoginAccessRequest.identifier] = {
        raw_packet: encoded,
        secret: LoginAccessRequest.secret
    };

    var RadiusClient = dgram.createSocket({ type: "udp4", reuseAddr: true });
    RadiusClient.bind(49001);
    var UseTimeoutExit = true;

    Logger.info("Users.RadiusLogin: Authenticating User=" + login.User + " (encoded.length=" + encoded.length + ")");
    RadiusClient.send(encoded, 0, encoded.length, 1812, SummitAuthenticationConfiguration.RADIUS.Address, function (err, bytes) {
        if (err != null) {
            UseTimeoutExit = false;// We got an error, no need to timeout
            UseTimeoutExit = false;
            Response.Result = false;
            Response.Message = "Error sending to RADIUS server:  Response=" + bytes;
            Logger.error("Users.RadiusLogin: Error sending to RADIUS server: Response=" + bytes);
            RadiusClient.close();
            callback(Response);
        }
    });

    // Really fucking ugly way to handle this - wish I could do it more elegantly
    setTimeout(function () {
        if (UseTimeoutExit) {
            Response.Result = false;
            Response.Message = "Timed out connecting to RADIUS server";
            Logger.error("Users.RadiusLogin: Timed out connecting to RADIUS server (5000ms)");
            try {
                RadiusClient.close();
            }
            catch (e) {
                Logger.error("Users.RadiusLogin: Error closing RADIUS client socket");
            } finally {
                callback(Response);
            }
        }
    }, 5000);

    RadiusClient.on('error', function (err) {
        UseTimeoutExit = false;// We got an error, no need to timeout

        Response.Result = false;
        Response.Message = "Error with RADIUS login: Check RADIUS Server configuration";
        Logger.error("Users.RadiusLogin: Error with RADIUS login: Error=" + err);
        RadiusClient.close();
        callback(Response);
    });

    RadiusClient.on('message', function (msg, rinfo) {
        UseTimeoutExit = false;// We got a response, no need to timeout

        var response = radius.decode({ packet: msg, secret: SummitAuthenticationConfiguration.RADIUS.Secret });
        var request = RadiusPacket[response.identifier];

        // although it's a slight hassle to keep track of packets, it's a good idea to verify
        // responses to make sure you are talking to a server with the same shared secret
        var valid_response = radius.verify_response({
            response: msg,
            request: request.raw_packet,
            secret: request.secret
        });

        if (valid_response) {
            Logger.info('Users.RadiusLogin: Valid response ' + response.code + ' for packet id ' + response.identifier);
            // take some action based on response.code
            switch (response.code) {

                case "Access-Reject":
                    SummitLoginAttempts.unshift({ "time": loginTime, "status": "Credentials did not match", "user": login.User, "ip": ip });
                    if (SummitLoginAttempts.length > SUMMIT_LOGIN_ATTEMPTS) SummitLoginAttempts.splice(-1);

                    Response.Result = false;
                    Response.Message = "Failed authentication";
                    break;

                case "Access-Accept":
                    SummitLoginAttempts.unshift({ "time": loginTime, "status": "Successful Login", "user": login.User, "ip": ip });
                    if (SummitLoginAttempts.length > SUMMIT_LOGIN_ATTEMPTS) SummitLoginAttempts.splice(-1);

                    Response.Result = true;
                    Response.Message = "Successful authentication";
                    break;
            }

            RadiusClient.close();
            callback(Response);
        } else {
            // don't take action since server cannot be trusted (but maybe alert user that shared secret may be incorrect)
            Logger.info('Users.RadiusLogin: Invalid response ' + response.code + ' for packet id ' + response.identifier);
            SummitLoginAttempts.unshift({ "time": loginTime, "status": "Internal Error", "user": login.User, "ip": ip });
            if (SummitLoginAttempts.length > SUMMIT_LOGIN_ATTEMPTS) SummitLoginAttempts.splice(-1);

            Response.Result = false;
            Response.Message = "Invalid response";
            RadiusClient.close();
            callback(Response);
        }
    });
}
//#endregion Login User

//#endregion Local Authentication

//#region Summit Authentication Configuration
exports.GetAuthenticationConfiguration = function (callback) {
    var summitAuthenticationConfiguration = JSON.parse(JSON.stringify(SummitAuthenticationConfiguration));
    //Logger.info("users.GetAuthenticationConfiguration: " + JSON.stringify(SummitAuthenticationConfiguration));
    callback(summitAuthenticationConfiguration);
}

// Put Summit Authentication Configuration
//#region Put Summit Authentication Configuration
exports.PutAuthenticationConfiguration = function (summitAuthenticationConfiguration, callback) {
    // Update SummitAuthenticationConfiguration
    SummitAuthenticationConfiguration = summitAuthenticationConfiguration;

    // Update users.json with new SummitLocalAuthenticationConfiguration settings to ensure persistence
    fs.writeFile('authentication.json', JSON.stringify(SummitAuthenticationConfiguration), function (error) {
        if (error) {
            Logger.error("users.PutAuthenticationConfiguration: Update write error. Error=" + error);
        }
        Logger.info("users.PutAuthenticationConfiguration: Successfully updated="+JSON.stringify(SummitAuthenticationConfiguration));
        callback(SummitAuthenticationConfiguration);
    });
}
//#endregion Put Summit Authentication Configuration

//#endregion Summit Authentication Configuration
//#endregion Authentication: Local and RADIUS

// HTTP/HTTPS Access
//#region HTTP/HTTPS Access

exports.ReadSummitSecurityOptions = function () {
    // Try reading security options, return as options JSON
    var summitprivatekey, summitcertificate;
    var options = {};

    try {
        summitprivatekey = fs.readFileSync('encryption/summit.key');
        summitcertificate = fs.readFileSync('encryption/summit.crt');
        options = {
            key: summitprivatekey,     // private key
            cert: summitcertificate    // self-signed certificate
                                       // intermediate certificate - NOT USED
        }
        Logger.info("Users.ReadSummitSecurityOptions: Successfully initialized Summit security options");
    }
    catch (e) {
        Logger.error("Users.ReadSummitSecurityOptions: Error initializing Summit security options=" + e);
    } finally {
        return options;
    }
}

//#region Summit Web Server Configuration

// This function will check every 500ms to see if web server configuration is loaded,
// If, after 5 attempts or 2500ms, it isn't loaded, it will send a message indicating failure (false)
// If, however, within 2500ms, the web server config is loaded, it will send a messaging indicating success (true)
var CheckIfLoadedCounter = 0;
var CheckIfLoadedTimer = null;
function WaitUntilWebServerConfigLoaded(callback) {
    CheckIfLoadedTimer = setInterval(function () {
        CheckIfLoadedCounter++;
        if (SummitWebServerConfiguration == null) {
            Logger.warn("Users.WaitUntilWebServerConfigLoaded: Web Server Config has not been initialized - attempt=" + CheckIfLoadedCounter);
            if (CheckIfLoadedCounter > 5) {
                clearInterval(CheckIfLoadedTimer);
                callback(false);
            }
        } else {
            Logger.warn("Users.WaitUntilWebServerConfigLoaded: Web Server Config has been initialized - attempt=" + CheckIfLoadedCounter);
            WebServerConfigLoaded = true;
            clearInterval(CheckIfLoadedTimer);
            callback(true);
        }
    }, 500);
}

exports.GetWebServerConfiguration = function (callback) {
    WaitUntilWebServerConfigLoaded(function (flagWebServerConfigLoaded) {
        // Depending on whether data was loaded, send back the Web Server config from local or factory settings
        if (flagWebServerConfigLoaded) {
            // Finally the web server configuration was initialized, let's read it
            Logger.warn("Users.GetWebServerConfiguration: Copying Web Server Config from previous settings=" + JSON.stringify(SummitWebServerConfiguration));
            var summitWebServerConfiguration = JSON.parse(JSON.stringify(SummitWebServerConfiguration));
            //Logger.info("users.GetWebServerConfiguration: " + JSON.stringify(SummitAuthenticationConfiguration));
            callback(summitWebServerConfiguration);
        } else {
            // Something went way wrong, we'll have to read it from the factory default
            Logger.warn("Users.GetWebServerConfiguration: Reading Web Server Config from default=" + JSON.stringify(SummitWebServerConfiguration));
            SummitWebServerConfiguration = require("./webserver.json");
            var summitWebServerConfiguration = JSON.parse(JSON.stringify(SummitWebServerConfiguration));
            callback(summitWebServerConfiguration);
        }
    });
}

// Put Summit Authentication Configuration
//#region Put Summit Authentication Configuration
exports.PutWebServerConfiguration = function (summitWebServerConfiguration, callback) {
    // Update SummitWebServerConfiguration
    SummitWebServerConfiguration = summitWebServerConfiguration;

    // Update users.json with new SummitWebServerConfiguration settings to ensure persistence
    fs.writeFile('webserver.json', JSON.stringify(SummitWebServerConfiguration), function (error) {
        if (error) {
            Logger.error("users.PutWebServerConfiguration: Update write error. Error=" + error);
        }
        Logger.info("users.PutWebServerConfiguration: Successfully updated=" + JSON.stringify(SummitWebServerConfiguration));
        callback(SummitWebServerConfiguration);
    });
}
//#endregion Put Summit Authentication Configuration

//#endregion Summit Authentication Configuration

//#endregion HTTP/HTTPS Access